#include "audio.h"
#include <string.h>
#include <linux/soundcard.h>
#include <fcntl.h>
#include <sys/ioctl.h>
#include <math.h>
#include <fstream>
using namespace std;

extern int SAMPLE_RATE;

audio_out::audio_out () {
	
	fd = -1;
	playing = 0;
	tid = 0;

	extern string dotdin;
	prefs_name = dotdin + "audio_prefs";
	
	load_prefs ();

}

void audio_out::load_prefs () {

	ifstream file (prefs_name.c_str(), ios::in);
	string ignore;
	
	if (!file) 
		defaults ();
	else {
		string type; 
		file >> ignore >> type;
		if (type != "audio_prefs") {
			cout << "bad audio prefrences file, please check: " << prefs_name << ", will use defaults." << eol;
			defaults ();
		} else {
			cout << "<< loading audio prefs from: " << prefs_name << eol;
			file >> ignore >> dsp_name; 
			file >> ignore >> sample_format;
			file >> ignore >> num_channels;
			file >> ignore >> SAMPLE_RATE; sample_rate = SAMPLE_RATE;
			file >> ignore >> fragment_power;
			file >> ignore >> fragment_select;
			cout << "+++ loaded audio prefs +++ " << eol;
		}
	} 	
	open ();
	close ();
}

void audio_out::save_prefs () {
	ofstream file (prefs_name.c_str(), ios::out);
	if (file) {
		file << "type: audio_prefs" << eol;
		file << "dsp_name: " << dsp_name << eol;
		file << "sample_format: " << sample_format << eol;
		file << "num_channels: " << num_channels << eol;
		file << "sample_rate: " << sample_rate << eol;
		file << "fragment_power: " << fragment_power << eol;
		file << "fragment_select: " << fragment_select << eol;
	} else cout << "couldnt write file: " << prefs_name << eol;
}

void audio_out::defaults () {
	
	cout << "setting up audio defaults..." << eol;
	
	dsp_name = "/dev/dsp";
	sample_format = AFMT_S16_NE;
	num_channels = 2; // stereo
	SAMPLE_RATE = sample_rate = 44100;

	// for more info see:
	// audio timing considerations: http://manuals.opensound.com/developer/audio_timing.html
	// SNDCTL_DSP_SETFRAGMENT: http://manuals.opensound.com/developer/SNDCTL_DSP_SETFRAGMENT.html
	fragment_power = 2;
	fragment_select = 8;
	
	cout << "+++ setup audio defaults +++" << eol;
	
}

int audio_out::open() {
	
	close ();
	
	fd = ::open (dsp_name.c_str(), O_WRONLY);
	if (fd == -1) {
		cons << console::red << "!!! couldnt open audio on: " << dsp_name << "!!!" << eol;
		return 0;
	} else cout << "+++ opened audio on: " << dsp_name << " +++" << eol;
	
	int fmt = sample_format;
  if (ioctl (fd, SNDCTL_DSP_SETFMT, &fmt) == -1) {
		cout << "!!! couldnt set the sample format. !!!" << eol;
  } 
  
  if (fmt != sample_format) {
		cons << console::red << "!!! sound card doesnt support your sample format. check " << prefs_name << " !!!" << eol;
		return 0;
	} else cout << "+++ sample format set +++" << eol;

	int nc = num_channels; 
  if (ioctl (fd, SNDCTL_DSP_CHANNELS, &nc) == -1) {
		cout << "!!! couldnt set number of channels: (SNDCTL_DSP_CHANNELS). !!!" << eol;
  } 
  
  if (nc != num_channels) {
		cons << console::red << "!!! cant set " << num_channels << " channels. check: " << prefs_name << " !!!" << eol;
		return 0;
	} else cout << "+++ channels set +++" << eol;

	int sr = sample_rate;
  if (ioctl (fd, SNDCTL_DSP_SPEED, &sr) == -1) {
		cons << console::red << "!!! couldnt set sample rate: " << sr << " !!!" << eol;
		return 0;
  } else {
		SAMPLE_RATE = sample_rate = sr;
		cout << "+++ sample rate: " << SAMPLE_RATE << " +++" << eol;
  }
  
	audio_buf_info bi;
	if (ioctl(fd, SNDCTL_DSP_GETOSPACE, &bi) == -1) {
		cout << "!!! couldnt get audio buffer info !!! " << eol;
	} else {
		cout << "<<< fragments info: " << "fragment_power: " << bi.fragments << ", fragment_size: " << bi.fragsize << " bytes, total: " << bi.fragstotal << ", bytes b4 block: " << bi.bytes << eol;
	}
	
	cout << "*** changing fragments info based on audio prefs ***" << eol;
	int frag = (fragment_power << 16) | fragment_select;
	ioctl(fd, SNDCTL_DSP_SETFRAGMENT, &frag);
	
	ioctl(fd, SNDCTL_DSP_GETOSPACE, &bi);
	cout << "<<< fragments info: " << "fragment_power: " << bi.fragments << ", fragment_size: " << bi.fragsize << " bytes, total: " << bi.fragstotal << ", bytes b4 block: " << bi.bytes << eol;
	  
  int trig = PCM_ENABLE_OUTPUT;
	if (ioctl(fd, SNDCTL_DSP_SETTRIGGER, &trig) == -1) {
		cons << console::red << "!!! couldnt trigger optimised output !!!" << eol;
	} 
	
	return 1;
			
}

int audio_out::close () {
	if (fd != -1) {
		if (::close (fd) == 0) {
			cout << "--- closed audio ---" << eol; 
			fd = -1;
			return 1;
		} else {
			cons << console::red << "!!! failed to close audio !!!" << eol;
		}
	}
	return 0;
}

audio_out::~audio_out () {
	stop ();
	save_prefs ();
}

int audio_out::start () {
	if (open ()) {
		if (fd > -1) {
			playing = 1;
			if (pthread_create (&tid, 0, audio_wanted, 0) != 0) {
				cons << console::red << "!!! could not stream audio !!! " << eol;
				return 0;
			} else {
				struct sched_param param;
				int policy;
				pthread_getschedparam(tid, &policy, &param);
				param.sched_priority = -20;
				pthread_setschedparam (tid, SCHED_FIFO, &param);
				pthread_getschedparam(tid, &policy, &param);
				cout << "<<< audio thread priority: " << param.sched_priority << eol;
				cout << "+++ streaming audio now +++" << eol;
				return 1;
			}
		}
	}
	return 0;
}

int audio_out::stop () {  
  if (fd > -1) {
		playing = 0;
		if (tid) pthread_join (tid, 0);
		tid = 0;
		return close();
	}
	return 0;
}

void audio_out::write () {
		
	for (int i = 0; i < SAMPLES_PER_BUFFER; ++i) {
		ibuf [i] = (short) (MAX_SHORT * fbuf [i]);
	}
	
	::write (fd, ibuf, IBUFFER_SIZE);
}

console& operator<< (console& c, const audio_out& ao) {
	c << console::yellow << "dsp_name: " << ao.dsp_name << " num_channels: " << ao.num_channels << " sample rate: " << ao.sample_rate << " sample_format: " << ao.sample_format << " fragment_power: " << ao.fragment_power << " fragment_select: " << ao.fragment_select;
	return c;
}
